@@ -211,7 +211,9 @@ def spa_worker(config, q_spa, pauseonload):
             done.append(action)
             if action == 'objstore':
                 altrepo = grokmirror.get_altrepo(fullpath)
-                grokmirror.fetch_objstore_repo(altrepo, fullpath)
+                # Should we use plumbing for this?
+                use_plumbing = config['core'].getboolean('objstore_uses_plumbing', False)
+                grokmirror.fetch_objstore_repo(altrepo, fullpath, use_plumbing=use_plumbing)
 
             elif action == 'repack':
                 logger.debug('quick-repacking %s', fullpath)
@@ -237,37 +239,43 @@ def spa_worker(config, q_spa, pauseonload):
 
 
 def objstore_repo_preload(config, obstrepo):
-    bname = os.path.basename(obstrepo)[:-4]
-    obstdir = os.path.realpath(config['core'].get('objstore'))
     purl = config['remote'].get('preload_bundle_url')
     if not purl:
-        # wing it -- it only costs us a single http request check
-        site = config['remote'].get('site')
-        if site and site.startswith('http'):
-            purl = '%s/objstore/preload' % site.rstrip('/')
-        else:
-            return
+        return
+    bname = os.path.basename(obstrepo)[:-4]
+    obstdir = os.path.realpath(config['core'].get('objstore'))
     burl = '%s/%s.bundle' % (purl.rstrip('/'), bname)
     bfile = os.path.join(obstdir, '%s.bundle' % bname)
-    sess = grokmirror.get_requests_session()
-    resp = sess.get(burl, stream=True)
-    resp.raise_for_status()
-    logger.info(' objstore: getting preload bundle for %s', bname)
-    with open(bfile, 'wb') as fh:
-        for chunk in resp.iter_content(chunk_size=8192):
-            fh.write(chunk)
-    resp.close()
+    try:
+        sess = grokmirror.get_requests_session()
+        resp = sess.get(burl, stream=True)
+        resp.raise_for_status()
+        logger.info(' objstore: downloading %s.bundle', bname)
+        with open(bfile, 'wb') as fh:
+            for chunk in resp.iter_content(chunk_size=8192):
+                fh.write(chunk)
+        resp.close()
+    except: # noqa
+        # Make sure we don't leave .bundle files lying around
+        # Should we add logic to resume downloads here in the future?
+        if os.path.exists(bfile):
+            os.unlink(bfile)
+        return
 
     # Now we clone from it into the objstore repo
     ecode, out, err = grokmirror.run_git_command(obstrepo, ['remote', 'add', '--mirror=fetch', '_preload', bfile])
     if ecode == 0:
-        logger.info(' objstore: preloading %s from the bundle', bname)
+        logger.info(' objstore: preloading %s.bundle', bname)
         args = ['remote', 'update', '_preload']
         ecode, out, err = grokmirror.run_git_command(obstrepo, args)
         if ecode > 0:
-            logger.info(' objstore: not able to preload, will clone repo-by-repo')
+            logger.info(' objstore: failed to preload from %s.bundle', bname)
         else:
-            logger.info(' objstore: successful preload')
+            # now pack refs and generate a commit graph
+            grokmirror.run_git_command(obstrepo, ['pack-refs', '--all'])
+            if grokmirror.git_newer_than('2.18.0'):
+                grokmirror.run_git_command(obstrepo, ['commit-graph', 'write'])
+            logger.info(' objstore: successful preload from %s.bundle', bname)
     # Regardless of what happened, we remove _preload and the bundle, then move on
     grokmirror.run_git_command(obstrepo, ['remote', 'rm', '_preload'])
     os.unlink(bfile)
@@ -279,6 +287,8 @@ def pull_worker(config, q_pull, q_spa, q_done):
     maxretries = config['pull'].getint('retries', 3)
     site = config['remote'].get('site')
     remotename = config['pull'].get('remotename', '_grokmirror')
+    # Should we use plumbing for objstore operations?
+    objstore_uses_plumbing = config['core'].getboolean('objstore_uses_plumbing', False)
 
     while True:
         try:
@@ -354,10 +364,7 @@ def pull_worker(config, q_pull, q_spa, q_done):
                 o_obj_info = grokmirror.get_repo_obj_info(obstrepo)
                 if o_obj_info.get('count') == '0' and o_obj_info.get('in-pack') == '0' and not my_fp:
                     # Try to preload the objstore repo directly
-                    try:
-                        objstore_repo_preload(config, obstrepo)
-                    except: # noqa
-                        pass
+                    objstore_repo_preload(config, obstrepo)
 
             if r_fp != my_fp:
                 # Make sure we have the remote set up
@@ -388,12 +395,13 @@ def pull_worker(config, q_pull, q_spa, q_done):
                             if o_obj_info.get('count') == '0' and o_obj_info.get('in-pack') == '0':
                                 # We fetch right now, as other repos may be waiting on these objects
                                 logger.info(' objstore: %s', gitdir)
-                                grokmirror.fetch_objstore_repo(altrepo, fullpath)
-                                spa_actions.append('repack')
+                                grokmirror.fetch_objstore_repo(altrepo, fullpath, use_plumbing=objstore_uses_plumbing)
+                                if not objstore_uses_plumbing:
+                                    spa_actions.append('repack')
                             else:
                                 # We lazy-fetch in the spa
                                 spa_actions.append('objstore')
-                                if my_fp is None:
+                                if my_fp is None and not objstore_uses_plumbing:
                                     # Initial clone, trigger a repack after objstore
                                     spa_actions.append('repack')
 
@@ -1237,7 +1245,7 @@ def pull_mirror(config, nomtime=False, forcepurge=False, runonce=False):
                                         return 0
                                     time.sleep(1)
                     # Don't run a hot loop waiting on results
-                    time.sleep(1)
+                    time.sleep(5)
                 continue
 
             if repoinfo is None:
@@ -1251,8 +1259,9 @@ def pull_mirror(config, nomtime=False, forcepurge=False, runonce=False):
                 if loopmark is None:
                     loopmark = gitdir
                 elif loopmark == gitdir:
-                    # sleep for a bit if we recognize that we've looped around all waiting repos
-                    time.sleep(0.5)
+                    # We've looped around all waiting repos, so back off and don't run
+                    # a hot waiting loop.
+                    time.sleep(5)
                 continue
 
             if gitdir == loopmark:
